"""
AutoAssociation.py
Auto-association or attractor network demonstration program

Please do not redistribute this code or modifications of it without the following notice:
(c) E T Rolls www.oxcns.org 2015
The background to these networks is provided in
Rolls ET (2021) Brain Computations: A Systems-Level Analysis (Oxford University Press) Appendix 2 (B)
which is available at https://www.oxcns.org by following the link to the book.
The instructions and exercises for this code are in
Rolls ET (2021) Brain Computations: A Systems-Level Analysis (Oxford University Press) Appendix 4 (D)
which is available at https://www.oxcns.org by following the link to the book.

This program is a translation into Python of the Matlab / Octave program AutoAssociation.m.
This program is intended to demonstrate some of the principles of how autoassociation or attractor networks operate,
and for that reason code optimizations are frequently eschewed in favour of illustrating principles of neuronal operation.
This program therefore also uses as few external libraries.

Note that if the program stops, it may be waiting for keyboard input.

The synaptic weight matrix (SynMat) has column vectors to represent the synapses on a single neuron.
There is one column vector therefore for each of the N neurons.
Thus SynMat[Row, Column] implements SynMat[synapse, neuron]
"""

# import necessary libraries
import numpy as np
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings("ignore")               # do not show unnecessary figure axis warnings


fig = 1                                         # index of plotting figures
N = 100                                         # number of neurons in the fully connected autoassociation net
nSyn = N                                        # number of synapses on each neuron
SynMat = np.zeros((nSyn, N))                    # Synapse weight matrix
nPatts = 10                                     # the number of training and testinng patterns. Suggest 10 for sparseness = 0.5, and 30 for sparseness = 0.
Learnrate = 1 / nPatts                          # not of special significance in an autoassociation network, but this keeps the weights within a range
Sparseness = 0.5                                # the sparseness of the trained representation, i.e., the proportion of neurons that have high firing rate of 1
                                                # Investigate values of 0.5 and 0.1
display = 1                                     # 0 no display; 1 display the network
nFlipBits = 14                                  # The number of bits that are flipped to produce distorted recall cues.
                                                # It is suggested that this be set to close to 0.2 * N * sparseness
nepochs = 9                                     # the number of times that the network is allowed to update during testing


# generate some binary training patterns
TrainPatts = np.zeros((N, nPatts))              # this matrix stores the traiining patterns. Each pattern vector has N elements
for patt in range(nPatts):
     TrainPatts[:int(N*Sparseness),patt] = 1  # the number of bits set to 1 in each pattern
     p = np.random.permutation(range(N))        # rearrange the elements of this pattern vector in random order
     TrainPatts[:,patt] = TrainPatts[p,patt]

if display == 1:
    fig += 1
    plt.title('Random Binary Training Patterns. Each column is a pattern')
    plt.xlabel('Pattern Number')
    plt.ylabel('Training Pattern')
    plt.imshow(TrainPatts, cmap='gray', aspect='auto')
    #plt.show(block=False)
    plt.pause(0.1)

input('Press any key to flip some of bits the patterns: ')
# Now make some testing patterns with some bits flipped from the training Patterns
TrainPattsFlipped = np.copy(TrainPatts)
for patt in range(nPatts):
    synarray = np.random.permutation(range(nSyn))
    el = 0
    for bit in range(nFlipBits):
        while TrainPatts[synarray[el],patt] != 1:
            el += 1
            if el > nSyn:
                print('Error: too many bits being flipped')
                el = 1
        TrainPattsFlipped[synarray[el],patt] = 0
        el += 1
    synarray = np.random.permutation(range(nSyn))
    el = 0
    for bit in range(nFlipBits):
        while TrainPatts[synarray[el],patt] != 0:
            el += 1
            if el > nSyn:
                print('Error: too many bits being flipped')
                el = 1
        TrainPattsFlipped[synarray[el],patt] = 1
        el += 1

if display == 1:
    fig += 1
    plt.figure(fig)
    plt.title('Training patterns with ' + str(nFlipBits) + ' flipped bits for testing')
    plt.xlabel('Pattern number')
    plt.ylabel('Distorted Test Pattern')
    plt.imshow(TrainPattsFlipped, cmap='gray', aspect='auto')
    #plt.show(block=False)
    plt.pause(0.1)

input('Press any key to train the auto association network: ')
# Train the synaptic matrix
for patt in range(nPatts):
    for neuron in range(N):
        postSynRate = TrainPatts[neuron,patt] # postsynaptic firing rate. The external input to the neurons
        for syn in range(nSyn):
            if syn != neuron: # avoid self connections of a recurrent collateral axon onto its sending neuron
                preSynRate = TrainPatts[syn,patt] # the presynaptic rate is the same as the postsynaptic rate because of the recurrent collaterals
                weight_change = Learnrate * (postSynRate - Sparseness) * (preSynRate - Sparseness) # use a covariance rule
                # the sparseness is the average firing rate
                # weight_change = Learnrate * postSynRate * preSynRate # or use a Hebb rule
                # weight_change = Learnrate * postSynRate * (preSynRate - Sparseness) # or use a Hebb rule but with also heterosynaptic LTD see Rolls (2008) B.3.3.6.
                SynMat[syn,neuron] += weight_change

    if display == 1:
        plt.figure(100)
        ax2 = plt.axes(title='Synaptic Matrix during learning pattern ' + str(patt+1), position=[0.2, 0.2, 0.7, 0.7])
        ax2.imshow(SynMat, cmap='gray', aspect = 'auto')
        ax1 = plt.axes(title='Presynaptic Input', position=[0.1, 0.2, 0.03, 0.7])
        ax1.imshow(np.reshape((TrainPatts[:,patt]), (-1,1)), cmap='gray', aspect='auto')
        ax1.set_xticklabels([])
        ax4 = plt.axes(title='Postsynaptic Rate', position=[0.2, 0.05, 0.7, 0.05])
        ax4.imshow(np.transpose(np.reshape(TrainPatts[:,patt], (-1,1))), cmap='gray', aspect='auto')
        ax4.set_yticklabels([])
        plt.pause(0.1)
        plt.draw()
        if patt < (nPatts - 1):
            plt.clf()

# SynMat
# print('Synaptic matrix minimum =', str(np.min(SynMat)), 'maximum =', str(np.max(SynMat)))
print(f'Synaptic matrix minimum = {np.min(SynMat):.03f}  maximum = {np.max(SynMat):.03f}')
# comment in the following two lines to test whether the network operates with positive-only synaptic weight
# SynMat -= np.min(SynMat)
# print('Synaptic matrix minimum= ', str(np.min(SynMat), 'maximum =', str(np.max(SynMat))))


# Now test the network to show wether it completes incomplete patterns and corrects
# distortions in the applied recall cue
input('Press any key for Testing: ')
nCorrect = 0

for patt in range(nPatts):
    Rate = np.zeros(N)      # output firing rates
    for epoch in range(nepochs):
        if epoch <= 0:
            clamp = 1           # the recall cue or pattern is applied only at the begining of retrieval
        else:
            clamp = 0
        # calculate activation
        Actvn = np.zeros(N) # the activations of the N output neurons
        for neuron in range(N):
            PreSynInput = Rate + (TrainPattsFlipped[:,patt] * clamp)  # the firing on the recurrent collaterals is the firing rate of the neurons
                                                                        # plus at the begining of retrieval the cue recall pattern
            Actvn[neuron] = np.dot(PreSynInput, SynMat[:,neuron])
        scale = 1 / ((np.max(Actvn) - np.min(Actvn)))
        Actvn = (Actvn - np.min(Actvn)) * scale
        # Now convert the activation into a firing rate using a binary threshold activation function
        # The threshold is selected in this case artificially by sorting
        # The array to determine where the threshold should be set to achieve the required sparseness
        # In the brain, inhibitory neurons are important in setting the proportion of excitatory neurons that are active
        tmp = sorted(list(Actvn), reverse=True)
        el = np.floor(len(Rate) * Sparseness)
        Threshold = tmp[int(el)-1]
        Rate = np.copy(Actvn)
        for neuron in range(len(Rate)):
            if Rate[neuron] >= Threshold:
                Rate[neuron] = 1
            else:
                Rate[neuron] = 0

        if display == 1:
            plt.figure(101)
            ax2 = plt.axes(title='Synaptic Matrix', position=[0.2, 0.25, 0.7, 0.7])
            ax2.imshow(SynMat, cmap='gray', aspect='auto')
            ax3 = plt.axes(title='Activation', position=[0.2, 0.125, 0.7, 0.03])
            ax3.imshow(np.reshape(Actvn, (1, -1)), cmap='gray', aspect='auto')
            ax3.set_yticklabels([])
            ax4 = plt.axes(title='Postsynaptic Rate', position=[0.2, 0.0, 0.7, 0.03])
            ax4.imshow(np.reshape(Rate, (1, -1)), cmap='gray', aspect='auto')
            ax4.set_yticklabels([])
            ax1 = plt.axes(title='Presynaptic input', position=[0.1, 0.25, 0.02, 0.7])
            ax1.imshow(np.reshape(PreSynInput, (-1, 1)), cmap='gray', aspect='auto')
            ax1.set_xticklabels([])
            plt.pause(0.1)
            plt.draw()
            if patt < (nPatts - 1):
                plt.clf()

        R1 = np.corrcoef(TrainPatts[:,patt], Rate)
        R = R1[0, 1]
        print(f'Patt  {patt+1}   epoch {epoch+1}  Correlation of recalled firing rate vector with training pattern = {R:.03f}')


    if R > 0.98:        # the criterion of correct is a correlation of 0.98 between a recalled pattern and the stored pattern
        nCorrect += 1   # accumulate the number of correct recalls across patterns

PercentCorrect = nCorrect / nPatts * 100
print('Percentage correct recall = '+ str(PercentCorrect))
