"""
CompetitiveNetDemo.py
Competitive network demonstration program

Please do not redistribute this code or modifications of it without the following notice:
(c) E T Rolls www.oxcns.org 2015
The background to these networks is provided in
Rolls ET (2021) Brain Computations: A Systems-Level Analysis (Oxford University Press) Appendix 2 (B)
which is available at http://www.oxcns.org by following the link to the book.
The instructions and exercises for this code are in
Rolls ET (2021) Brain Computations: A Systems-Level Analysis (Oxford University Press) Appendix 4 (D)
which is available at http://www.oxcns.org by following the link to the book.

This program is a translation into P
ython of the Matlab / Octave program CompetitiveNetDemo.m.
This program is intended to demonstrate some of the principles of how competitive networks operate,
and for that reason code optimizations are frequently eschewed in favour of illustrating principles of neuronal operation.
This program therefore also uses few external libraries.

Note that if the program stops, it may be waiting for keyboard input.

The synaptic weight matrix (SynMat) has column vectors to represent the synapses on a single neuronn.
There is one column vector therefore for each of the N neurons.
Thus SynMat(Row, Column) implements SynMat(synapse, neuron)
"""

# import necessary libraries
import numpy as np
import matplotlib.pyplot as plt
from normVecLen import normVecLen
from spars import spars
import time
import warnings
warnings.filterwarnings("ignore")   # do not show unnecessary figure axis warnings


fig = 0
N = 100                             # number of neurons in the competitive net
nSyn = N                            # number of synapses on each neuron that are not deleted. Set the dilution by the divisor
nepochs = 10                        # number of training epochs, default 10
Learnrate = 0.1                     # ~ 0.1 works, Aim for weightchange = 0.02 - 0.2
if Learnrate == 0: nepochs = 0
OutputSparseness = 0.01             # the sparseness of output firing
display = 1                         # 0 no display; 1 display during learning epoch 1
Actvn = np.zeros(N)                 # activation
Rate = np.zeros(N)                  # output firing rates


# generating training patterns overlapping by shift
leng = 20                           # length of each training pattern
InputSparseness = leng / nSyn
shift = 3                           # shift between training patterns
nPatts = 30
TrainPatts = np.zeros((N, nPatts))

for patt in range(nPatts):
    for x in range(leng):
        syn = patt * shift + x
        syn %= nSyn
        TrainPatts[syn,patt] = 1

    if (patt * shift + leng) >= nSyn:
        break

nPatts = patt + 1   # this will set nPatts to be the number that can be fitted into the array without much wrap-around
print('Number of patterns generated = ' + str(nPatts) + ' with shift of '+ str(shift) + ' and length of ' + str(leng))

for patt in range(nPatts):
    TrainPatts[:, patt] /= np.linalg.norm(TrainPatts[:, patt], 2)       # normalize the length of each training pattern
TrainPatts = np.nan_to_num(TrainPatts)

if display == 1:
    fig += 1
    plt.figure(fig)
    plt.title('Training Patterns')
    plt.xlabel('Pattern Number')
    plt.ylabel('Training Pattern')
    plt.imshow(TrainPatts, cmap='gray', aspect='auto')
    #plt.show(block=False)
    plt.pause(1)

input('Press any key to Train the competitive network: ')

SynMat = np.random.rand(nSyn, N)        # (Synapses = Rows, neurons = Columns)
for neuron in range(N):
    SynMat[:, neuron] = normVecLen(SynMat[:, neuron])           # normalize the synaptic weight vector on each neuron


# train the synaptic matrix
MeanPreSynapticRate = np.sum(TrainPatts[:, 0]) / nSyn
for epoch in range(nepochs):
    order = np.random.permutation(nPatts)           # present the patterns in random order during training
    for i in range(nPatts):
        patt = order[i]
        #calculate activation
        for neuron in range(N):
            PreSynInput = TrainPatts[:, patt]                           # the retrieval cue is the conditioned stimulus
            Actvn[neuron] = np.dot(PreSynInput, SynMat[:,neuron])       # the dot product of input pattern and synaptic weight vector for one neuron
        scale = 1.0 / (np.max(Actvn) - np.min(Actvn))
        Actvn = (Actvn - np.min(Actvn)) * scale                         # scale activation to 0-1

        # now convert the activation into a firing rate using a binary or linear threshold activation function
        # The threshold is selected in this case artificially by sorting
        # The array to determine where the threshold should be set to achieve the required sparseness
        # In the brain, inhibitory neurons are import in setting the proportion of excitatory neurons that are active

        tmp = sorted(Actvn, reverse=True)
        el = np.floor(len(Rate) * OutputSparseness)
        Threshold = tmp[int(el-1)]
        Rate = np.copy(Actvn)
        for neuron in range(len(Rate)):
            if Rate[neuron] >= Threshold:                               # threshold binary or linear activation function
                Rate[neuron] = 1                                        # comment out this line for a threshold linear activation function
            else:
                Rate[neuron] = 0

        # now change the weights in the synaptic matrix
        for neuron in range(N):
            for syn in range(nSyn):
                preSynRate = TrainPatts[syn, patt]
                weight_change = Learnrate * Rate[neuron] * preSynRate
                # weight_change = Learnrate * Rate[neuron] * (preSynRate - MeanPreSynapticRate) # or try an associative rule with heterosynaptic LTD
                SynMat[syn,neuron] += weight_change
            SynMat[:, neuron] = normVecLen(SynMat[:, neuron])           # normalize the neuron's synaptic weight vector

        if display == 1:
            plt.figure(100)
            ax2 = plt.axes(title='Synaptic Matrix', position=[0.2, 0.25, 0.7, 0.7])
            ax2.imshow(SynMat, cmap='gray', aspect='auto')
            ax3 = plt.axes(title='Activation', position=[0.2, 0.125, 0.7, 0.03])
            ax3.imshow(np.reshape(Actvn, (1, -1)), cmap='gray', aspect='auto')
            ax3.set_yticklabels([])
            ax4 = plt.axes(title='Postsynaptic Rate', position=[0.2, 0.0, 0.7, 0.03])
            ax4.imshow(np.reshape(Rate, (1, -1)), cmap='gray', aspect='auto')
            ax4.set_yticklabels([])
            ax1 = plt.axes(title='Presynaptic input', position=[0.1, 0.25, 0.02, 0.7])
            ax1.imshow(np.reshape(PreSynInput, (-1, 1)), cmap='gray', aspect='auto')
            ax1.set_xticklabels([])
            plt.pause(0.1)
            plt.draw()
            plt.clf()
        print('Epoch', epoch+1, 'Pattern', patt + 1)



# Now Test it
input('Press Any Key to start Testing: ')
RecalledRates = np.zeros((N, nPatts))                   # the firing rate output vectors
sparsenessSum = 0
for patt in range(nPatts):
    # calculate activation
    Rate = np.zeros((1,N))                              # output firing rates
    for neuron in range(N):
        PreSynInput = TrainPatts[:, patt]               # the retrieval cue
        Actvn[neuron] = np.dot(PreSynInput, SynMat[:, neuron])      # dot product of input pattern and synaptic weight vector for one neuron
    scale = 1.0 / (np.max(Actvn) - np.min(Actvn))
    Actvn = (Actvn - np.min(Actvn)) * scale             # scale Activation to 0-1
    # Now convert the activation into a firing rate using a binary or linear threshold acticvation function
    # The threshold is selected in this case artificially by sorting
    # The array to determine where the threshold should be set to achieve the required sparseness
    # In the brain, inhibitory neurons are important in setting the proportion of excitatory neurons that are active
    tmp = sorted(Actvn, reverse=True)
    el = np.floor(len(Rate) * OutputSparseness)
    Threshold = tmp[int(el)]
    Rate = np.copy(Actvn)
    for neuron in range(N):
        if Rate[neuron] >= Threshold:                   # threshold binary or linear activation function
            Rate[neuron] = 1                            # comment out this line for a threshold linear activation function
        else:
            Rate[neuron] = 0
    RecalledRates[:, patt] = np.copy(Rate)

    if display == 1:
        plt.figure(101)
        ax2 = plt.axes(title='Synaptic Matrix', position=[0.2, 0.25, 0.7, 0.7])
        ax2.imshow(SynMat, cmap='gray', aspect='auto')
        ax3 = plt.axes(title='Activation', position=[0.2, 0.125, 0.7, 0.03])
        ax3.set_yticklabels([])
        ax3.imshow(np.reshape(Actvn, (1,-1)), cmap='gray', aspect='auto')
        ax4 = plt.axes(title='Postsynaptic Rate', position=[0.2, 0.0, 0.7, 0.03])
        ax4.set_yticklabels([])
        ax4.imshow(np.reshape(Rate, (1,-1)), cmap='gray', aspect='auto')
        ax1 = plt.axes(title='Presynaptic input', position=[0.1, 0.25, 0.02, 0.7])
        ax1.set_xticklabels([])
        ax1.imshow(np.reshape((PreSynInput), (-1,1)), cmap='gray', aspect='auto')
        sparse = spars(Rate)
        sparsenessSum += sparse
        #plt.show(block=False)
        plt.pause(0.1)
        input(f'Press any key to test with pattern {patt+2}: ')

MeanSparseness = sparsenessSum / nPatts
input('Press any key to show the Pattern vs. Pattern Correlation matrix: ')

# show correlation matrix between the different training patterns for comparison
PattCorMat = np.zeros((nPatts, nPatts))
for i in range(nPatts):
    for j in range(nPatts):
        R1 = np.corrcoef(TrainPatts[:, i], TrainPatts[:, j])
        R = R1[0, 1]
        PattCorMat[j, i] = R

fig += 1
plt.figure(fig)
x = np.linspace(-1, 1, 21)
plt.title(f'Correlations between input patterns {np.min(PattCorMat):.03f} to {np.max(PattCorMat):.03f}')
plt.xlabel('Pattern Number')
plt.ylabel('Pattern Number')
plt.imshow(PattCorMat, cmap='gray', aspect='auto')
cbar = plt.colorbar()
cbar.set_label('correlation', rotation=90)
#plt.show(block=False)
plt.pause(0.1)

# and get the mean abs value of the off-diagonal terms, which are the correlations between
# the different patterns
numvals = 0
PattsumCor = 0
for col in range(nPatts):
    for row in range(nPatts):
        if row != col:
            PattsumCor += abs(PattCorMat[row, col])
            numvals += 1
print(f'Average of absolute values of correlation between different input patterns= {(PattsumCor / numvals):.03f}')


input('Press any key to show the Output Rates Correlation matrix: ')
# show correlation matrix between the output rates for the different training patterns
CorMat = np.zeros((nPatts, nPatts))
for i in range(nPatts):
    for j in range(nPatts):
        R1 = np.corrcoef(RecalledRates[:, i], RecalledRates[:, j])
        R = R1[0, 1]
        CorMat[j, i] = R


# and get the mean abs value of the off-diagonal terms, which are the
# correlations between the rates output for the different patterns
fig += 1
plt.figure(fig)
x = np.linspace(-1, 1, 21)
plt.title(f'Correlations between output Rates {np.min(PattCorMat):.03f} to {np.max(PattCorMat):.03f}')
plt.xlabel('Pattern Number')
plt.ylabel('Pattern Number')
plt.imshow(CorMat, cmap='gray', aspect='auto')
cbar = plt.colorbar()
cbar.set_label('correlation', rotation=90)
plt.show(block=False)
plt.pause(3)
#plt.close('all')

numvals = 0
sumCor = 0
for col in range(nPatts):
    for row in range(nPatts):
        if row != col:
            sumCor += abs(CorMat[row, col])
            numvals += 1
print(f'Average of absolute values of correlation between output rates to different patterns= {(sumCor / numvals):.03f}')
